function [W,analysis_data,cost,isi] = iva_laplace(X,varargin)
% [W,cost,isi] = iva_laplace(X,varargin)
analysis_data = [];
% Implementation of the independent vector analysis (IVA)
% algorithm using an uncorrelated multivariate Laplace prior.  This results
% in the same score function as used in the original IVA (see references
% below).
%
% Input:
% X - data observations from K data sets, i.e. X{k}=A{k}*S{k}, where A{k}
% is an N x N unknowned invertible mixing matrix and S{k} is N x T  matrix
% with the nth row corresponding to T samples of the nth source in the kth
% dataset.  For IVA it is assumed that the source is statistically
% independent of all the sources within a dataset and exactly dependent on
% at most one source in each of the other datasets. The data, X, is a
% 3-dimensional matrix of dimensions N x K x T.  The latter enforces the
% assumption of an equal number of samples in each dataset.
%
% Output:
% W - the estimated demixing matrices so that ideally W{k}*A{k} = P*D{k}
% where P is any arbitrary permutation matrix and D{k} is any diagonal
% invertible (scaling) matrix.  Note P is common to all datasets; this is
% to indicate that the local permuation ambiguity between dependent sources
% across datasets should ideally be resolved by IVA.
%
% cost - the cost for each iteration
%
% isi - joint inter-symbol-interference is available if user supplies true
% mixing matrices for computing a performance metric
%
% Optional input pairs and default values are given in list below:
%
%    'whiten',true, ... % whitening is optional
%    'verbose',false, ... % verbose true enables print statements
%    'A',[], ... % true mixing matrices A, automatically sets verbose
%    'W_init',[], ... % initial estimates for demixing matrices in W
%    'maxIter',2*512, ... % max number of iterations
%    'terminationCriterion','ChangeInCost', ... % criterion for terminating iterations: (ChangeInCost), ChangeInW
%    'termThreshold',1e-6, ... % termination threshold
%    'alpha0',0.1 ... % initial step size scaling
%
% Example call:
% W=iva_laplace(X)
%
% To test code call function with no input arguments: iva_laplace
%
% Coded by Matthew Anderson (matt.anderson at umbc.edu)
%
% References:
%
% [1] T. Kim, I. Lee, & T.-W. Lee, "Independent Vector Analysis: Definition and Algorithms," Proc. of 40th Asilomar Conference on Signals, Systems, and Computers, 2006, 1393-1396
% [2] T. Kim, T. Eltoft, & T.-W. Lee, "Independent Vector Analysis: an extension of ICA to multivariate components," Lecture Notes in Computer Science: Independent Component Analysis and Blind Signal Separation, Independent Component Analysis and Blind Signal Separation, Springer Berlin / Heidelberg, 2006, 3889, 165-172
% [3] T. Kim, H. T. Attias, S.-Y. Lee, & T.-W. Lee, "Blind Source Separation Exploiting Higher-Order Frequency Dependencies," IEEE Trans. Audio Speech Lang. Process., 2007, 15, 70-79

% Version 01 - 20120913 - Initial publication.
% Version 02 - 20120919 - Corrected help comments above and references.

%% Call internal test function when no arguments supplied
if nargin==0
   help iva_laplace
   test_iva_laplace
   return
end

%% Gather Options

% build default options structure
Params=struct( ...
   'whiten',true, ... % whitening is optional
   'verbose',false, ... % verbose true enables print statements
   'A',[], ... % true mixing matrices A, automatically sets verbose
   'initW',[], ... % initial estimates for demixing matrices in W
   'maxIter',2*512, ... % max number of iterations
   'terminationCriterion','ChangeInCost', ... % criterion for terminating iterations: (ChangeInCost), ChangeInW
   'termThreshold',1e-6, ... % termination threshold
   'alpha0',0.1 ... % initial step size scaling
   );

% load in user supplied options
Params=getopt(Params,varargin{:});
alphaMin=0.1; % alpha0 will max(alphaMin,alphaScale*alpha0)
alphaScale=0.9;
supplyA=~isempty(Params.A); % set to true if user has supplied true mixing matrices
outputISI=false;

%% Create cell versions
% Faster than using the more natural 3-dimensional version
if iscell(Params.A)
   K=length(Params.A);
   A=zeros([size(Params.A{1}) K]);
   for k=1:K
      A(:,:,k)=Params.A{k};
   end
   Params.A=A;
   clear A
end

[N,T,K]=size(X); % the input format insures all datasets have equal number of samples

if Params.whiten
   [X,V]=whiten(X);
end


%% Initialize W
if ~isempty(Params.initW)
   W=Params.initW;
   if size(W,3)==1 && size(W,1)==N && size(W,2)==N
      W=repmat(W,[1,1,K]);
   end
   if Params.whiten
      for k=1:K
         W(:,:,k)=W(:,:,k)/V(:,:,k);
      end
   end
else
   W=randn(N,N,K);
end

%% When mixing matrix A supplied
% verbose is set to true
% outputISI can be computed if requested
% A matrix is conditioned by V if data is whitened
if supplyA
   % only reason to supply A matrices is to display running performance
   Params.verbose=true;
   if nargout>2
      outputISI=true;
      isi=nan(1,Params.maxIter);
   end
   if Params.whiten
      for k = 1:K
         Params.A(:,:,k) = V(:,:,k)*Params.A(:,:,k);
      end
   end
end

%% Initialize some local variables
cost=nan(1,Params.maxIter);

Y=X*0;

%% Main Iteration Loop
for iter = 1:Params.maxIter
   termCriterion=0;
   
   % Current estimated sources
   for k=1:K
      Y(:,:,k)=W(:,:,k)*X(:,:,k);
   end
   
   % Inner Product of N SCVs (used to compute score function
   sqrtYtY=sqrt(sum(abs(Y).^2,3));
   sqrtYtYInv=sqrtYtY.^(-1);
   
   % Some additional computations of performance via ISI when true A is supplied
   if supplyA
      [amari_avg_isi,amari_joint_isi]=bss_isi(W,Params.A);
      if outputISI
         isi(iter)=amari_joint_isi;
      end
   end
   
   W_old=W; % save current W as W_old
   cost(iter)=0;
   for k=1:K
      cost(iter)=cost(iter)+log(abs(det(W(:,:,k))));
   end
   cost(iter)=sum(sum(sqrtYtY))/T-cost(iter);
   cost(iter)=cost(iter)/(N*K);
   
   %% Loop over each dataset
   dW=W*0;
   for k=1:K
      % Calculate multivariate score function
      phi=sqrtYtYInv.*Y(:,:,k);
      dW(:,:,k)=W(:,:,k)-phi*Y(:,:,k)'/T*W(:,:,k);
   end
   
   %% Update W
   W=W+Params.alpha0*dW;
   
   %% Calculate termination criterion
   switch lower(Params.terminationCriterion)
      case lower('ChangeInW')
         for k=1:K
            termCriterion = max(termCriterion,max(1-abs(diag(W_old(:,:,k)*W(:,:,k)'))));
         end % k
      case lower('ChangeInCost')
         if iter==1
            termCriterion=1;
         else
            termCriterion=abs(cost(iter-1)-cost(iter))/abs(cost(iter));
         end
      otherwise
         error('Unknown termination method.')
   end
   
   %% Check the termination condition
   if termCriterion < Params.termThreshold || iter == Params.maxIter
      break;
   elseif isnan(cost(iter))
      for k = 1:K
         W(:,:,k) = eye(N) + 0.1*randn(N);
      end
      if Params.verbose
         fprintf('\n W blowup, restart with new initial value.');
      end
   elseif iter>1 && cost(iter)>cost(iter-1)
      % see if this improves convergence
      Params.alpha0=max(alphaMin,alphaScale*Params.alpha0);
   end
   
   %% Display Iteration Information
   if Params.verbose
      if supplyA
         fprintf('\n Step %d: W change: %f, Cost: %f, Avg ISI: %f, Joint ISI: %f',  ...
            iter, termCriterion,cost(iter),amari_avg_isi,amari_joint_isi);
                analysis_data = [analysis_data; iter, termCriterion,cost(iter),amari_avg_isi,amari_joint_isi];
      else
         fprintf('\n Step %d: W change: %f, Cost: %f', iter, termCriterion,cost(iter));
      end
   end % options.verbose
end % iter

%% Finish Display
if iter==1 && Params.verbose
   if supplyA
      fprintf('\n Step %d: W change: %f, Cost: %f, Avg ISI: %f, Joint ISI: %f',  ...
         iter, termCriterion,cost(iter),amari_avg_isi,amari_joint_isi);
   else
      fprintf('\n Step %d: W change: %f, Cost: %f', iter, termCriterion,cost(iter));
   end
end % options.verbose

if Params.verbose
   fprintf('\n');
end

%% Clean-up Outputs
if Params.whiten
   for k=1:K
      W(:,:,k)=W(:,:,k)*V(:,:,k);
   end
end


%% Return
return;
% END OF iva_laplace